/***************************************************************
* Name:         dispatch.c
*
* Purpose:      Simple command line text editor
 *              dispatcher functions
*
* Course code:  21022105
* Student name: Marius Joldos
* E-mail:       Marius.Joldos@cs.utcluj.ro
****************************************************************
* Created:      2008-11-07
*
****************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "ted.h"
#include "dispatch.h"
/***********************************************************
* FUNCTION NAME: getLineNumber
*
* DESCRIPTION:  gets a number from user and decrements it by one before returning it
*
*
* ARGUMENT LIST: none
* RETURN VALUE: int number read minus 1
*
* CHANGES: -
*************************************************************/
int getLineNumber(char *buffer)
{
    int lineNo = 0;
    if (1 != sscanf(buffer, "%d", &lineNo))
    {
        fprintf(stderr, "set: bad line number\n");
        return NONE;
    }
    if (lineNo < 0 || lineNo >= MAX_LINE)
    {
        fprintf(stderr, "set: line number outside limits: %d\n", lineNo);
        return NONE;
    }
    return --lineNo;
}

/***********************************************************
* FUNCTION NAME: getLastLine
*
* DESCRIPTION:  gets the line argument for commands which represents the
*               2nd arg to commands; returns the value of that line upon success,
*               or NONE if failure
*
* ARGUMENT LIST:
* Argument        Type   IO      Description
* ------------- -------- -- --------------------------------
* const char*   fctName  I    name of function invoking this one
* RETURN VALUE: int           value of last argument of command on success,
*                             or NONE on failure
*
* CHANGES: -
*************************************************************/
int getLastLine(char *buf, int lineNo, const char *fctName)
{
    char *ptrComma;
    int lastLine;
    if (!(ptrComma = strrchr(buf, ',')))
    {
        fprintf(stderr, "No manditory comma\n");
        return NONE;
    }
    lastLine = getLineNumber(++ptrComma);
    if (lastLine < lineNo)
    {
        fprintf(stderr, "Bad arguments to %s: %d, %d\n", fctName, lineNo, lastLine);
    }
    return lastLine;
}
/***********************************************************
* FUNCTION NAME: dispatch
*
* DESCRIPTION:  dispatcher for simple text editor
*                accepts user comamnds and invokes the proper
*                function in ted module
*
* ARGUMENT LIST: none
* RETURN VALUE: void
*
* CHANGES: fileName
*************************************************************/
int dispatch()
{
    char c; // command
    char buf[256]; // buffer for reading user command
    char *ptrComma; // pointer to last comma in buf
    char *arg; // auxiliary pointer for getting last line argument for commands using 2 line numbers
    int lineNo, firstLine, lastLine; // auxiliary variables for line numbers

    for (;;)
    {
        printf("Type h for help> "); // prompt for command
        fgets(buf, sizeof(buf), stdin); // read one line
        if (buf[strlen(buf)-1] == '\n') buf[strlen(buf)-1] = '\0';
        c = buf[0];
        switch (tolower(c))
        {
        case 'h':
            help();
            break;
        case 's': // set lineNumber
            lineNo = getLineNumber(&buf[1]);
            set(lineNo);
            break;
        case 'i': // insert text before current line
            insert();
            break;
        case 'a': // insert text after current line
            append();
            break;
        case 'd': // delete lines
            // d delete current
            // d firstline, lastLine delete all lines in the specified interval
            firstLine = getLineNumber(&buf[1]);
            if (firstLine == NONE) delete(crtLineIndex, crtLineIndex);
            else
            {
                if ((lastLine = getLastLine(buf, firstLine, "delete()")) != NONE)
                    delete(firstLine, lastLine);
            }
            break;
            // todo: the rest of the commands
            // ...
        case 'c': // copy
            // c copy current
            // c firstline, lastLine copies to scrap all lines in the specified interval
            firstLine = getLineNumber(&buf[1]);
            if (firstLine == NONE) copy(crtLineIndex, crtLineIndex);
            else
            {
                if ((lastLine = getLastLine(buf, firstLine, "copy()")) != NONE)
                    copy(firstLine, lastLine);
            }
            break;            break;
        case 'p': // paste
            // inserts scap buffer BEFORE the current line
            break;
        case 'v': // view
            firstLine = getLineNumber(&buf[1]);
            if (firstLine == -1) view(crtLineIndex, crtLineIndex);
            else
            {
                if ((lastLine = getLastLine(buf, firstLine, "view()")) != NONE)
                {
                    if (lastLine < lineNo)
                    {
                        fprintf(stderr, "Bad arguments to view: %d, %d\n", firstLine, lastLine);
                    }
                    else view(firstLine, lastLine);
                }
            }
            break;
        case 'k': //cut

            break;
        case 'q':
            return 1;
            break;
        case 'n':
            arg = strrchr(buf, ' ');
            // todo: check if arg is at least 1 char long
            strcpy(fileName, arg + 1);
            break;
        case 'l': // load
            arg = strrchr(buf, ' ');
            if (arg)
            {
                strcpy(fileName, arg + 1);
                load(fileName);
            }
            else if (*fileName)
            {
                load(fileName);
            }
            else
                fprintf(stderr, "Bad name for load: %s\n", fileName);
            break;
        case 'w': // write
            // todo: get file name
            arg = strrchr(buf, ' ');
            if (arg)
            {
                strcpy(fileName, arg + 1);
                write(fileName);
            }
            else if (*fileName)
            {
                write(fileName);
            }
            else
                fprintf(stderr, "Bad name for write: %s\n", fileName);
            break;
        case 'e': // exit and save changes
            if (strlen(fileName) > 0) write(fileName);
            else break;
            return 0;
        default:
            fprintf(stderr, "Unknowm command %c\n", c);
            break;
        }
    }
    return 0;
}
