/***************************************************************
* Name:         ted.c
*
* Purpose:      Simple command line text editor support
*
* Course code:  21022105
* Student name: Marius Joldos
* E-mail:       Marius.Joldos@cs.utcluj.ro
****************************************************************
* Created:      2008-11-04
*
****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ted.h"
// global variables
static char lines[MAX_LINE][MAX_LINE_LEN]; // array to store edited lines
static char scrap[MAX_LINE][MAX_LINE_LEN]; // array to use as scrap buffer for cut, paste and copy
static char buffer[MAX_LINE][MAX_LINE_LEN]; // buffer for input lines for insert and append
int crtLineIndex = 0; // current line index
int lastLineIndex = 0; // last filled line
static int lastBufferLine = NONE; // last filled line in buffer
static int lastScrapLine = NONE; // last filled line in scrap
char fileName[256]; // name of file to load or write
static char *helpText[] =
{
    "Simple text editor. Commands are:"
    "\th - show this help",
    "\ts - set current line",
    "\ti - insert user provided text before current line; end input with Ctrl/Z",
    "\ta - append user provided text after current line; end input with Ctrl/Z",
    "\td - delete",
    "\t\td with no parameters deletes current line",
    "\t\td lineNo : delete line lineNo",
    "\t\td from, to : delete all lines in interval [from, to]",
    "\tc - copy",
    "\t\tc with no parameters copies current line to scrap",
    "\t\tc lineNo : copy line lineNo",
    "\t\tc from, to : copy all lines in interval [from, to]",
    "\tp - paste : pastes scrap buffer before current line",
    "\tk - cut",
    "\t\tk with no parameters cuts current line to scrap",
    "\t\tk lineNo : cut line lineNo",
    "\t\tk from, to : cut all lines in interval [from, to]",
    "\tv - view",
    "\t\tv with no parameters lists current line",
    "\t\tv lineNo : list line lineNo",
    "\t\tv from, to : list all lines in interval [from, to]",
    "\tl fileName - load fileName into editor",
    "\tw fileName - write editor buffer to fileName. If file exists it is overwritten",
    "\tq - quit without saving changes",
    "\te - exit and write changes"
};

/***********************************************************
* FUNCTION NAME: moveUp
*
* DESCRIPTION: moves a number of size lines up in lines array
*              to make room
*
* ARGUMENT LIST:
* Argument        Type   IO      Description
* ------------- -------- -- --------------------------------
* int           fromIndex I first line to move
* int           toIndex   I destination line for 1st line
* int           size      I how many lines
* RETURN VALUE: static void
*
* CHANGES: lines
*************************************************************/
static void moveUp(const int fromIndex, const int toIndex, const int size)
{
    for (int i = size - 1; i >= 0; i--)
        strcpy(lines[fromIndex + i], lines[toIndex + i]);
}
/***********************************************************
* FUNCTION NAME: moveDown
*
* DESCRIPTION: moves a number of size lines down in lines array
*              to make room
*
* ARGUMENT LIST:
* Argument        Type   IO      Description
* ------------- -------- -- --------------------------------
* int           fromIndex I first line to move
* int           toIndex   I destination line for 1st line
* int           size      I how many lines
* RETURN VALUE: static void
*
* CHANGES: lines
*************************************************************/
static void moveDown(const int fromIndex, const int toIndex, const int size)
{
    for (int i = 0; i < size; i++)
        strcpy(lines[fromIndex + i], lines[toIndex + i]);
}


/***********************************************************
* FUNCTION NAME: getInput
*
* DESCRIPTION:  accepts input from the user up to EOF (Ctrl/Z)
*
*
* ARGUMENT LIST: none
* RETURN VALUE: static void
*
* CHANGES: buffer
*************************************************************/
static void getInput()
{
    int i = 0;

    while (fgets(buffer[i], sizeof(buffer[i]), stdin)) i++;
    lastBufferLine = i - 1;
}
/***********************************************************
* FUNCTION NAME: help
*
* DESCRIPTION:  show help info
*
*
* ARGUMENT LIST: none
* RETURN VALUE: void
*
* CHANGES: -
*************************************************************/
void help()
{
    int i;
    for (i = 0; i < sizeof(helpText)/sizeof(helpText[0]); i++)
        printf("%s\n", helpText[i]);
}
/***********************************************************
* FUNCTION NAME: set
*
* DESCRIPTION:  sets cursor at given line number
*
* ARGUMENT LIST:
* Argument        Type   IO      Description
* ------------- -------- -- --------------------------------
* int           lineNo    I line to set cursor to
* RETURN VALUE: void
*
* CHANGES:      global lineNO
*************************************************************/
void set(int lineNo)
{
    if (lineNo >=0 && lineNo < MAX_LINE)
        crtLineIndex = lineNo;
    else
        fprintf(stderr, "Bad arguiment to set %d\n", lineNo);
}
/***********************************************************
* FUNCTION NAME: insert
*
* DESCRIPTION:  accepts input from the user, ended with Ctrl/Z
*                inserts all this input after current line
*
* ARGUMENT LIST: none
* RETURN VALUE: int  zero if it fails, 1 if it succeeds
*
* CHANGES: buffer, lines
*************************************************************/
int insert()
{
    int success = FALSE;
    getInput(buffer); // get user input up to Ctrl/Z
    // todo: make room and insert buffer into lines before current line
    printf("insert(): not complete\n");
    return success;

}
/***********************************************************
* FUNCTION NAME: append
*
* DESCRIPTION:  accepts input from the user, ended with Ctrl/Z
*                inserts all this input before current line
*
* ARGUMENT LIST: none
* RETURN VALUE: int  zero if it fails, 1 if it succeeds
*
* CHANGES: buffer, lines
*************************************************************/
int append()
{
    int success = FALSE;
    getInput(buffer); // get user input up to Ctrl/Z
    // todo: make room and insert buffer into lines after current line
    printf("append(): not complete\n");
    return success;
}
/***********************************************************
* FUNCTION NAME: delete
*
* DESCRIPTION:  deletes all lines from beginLineIndex to emdLineIndex, inclusive
*
* ARGUMENT LIST:
* Argument        Type         IO      Description
* ------------- --------       -- --------------------------------
* int           beginLineIndex  I  index of 1st line to delete
* int           endLineIndex    I  index of last line to delete
* RETURN VALUE: int  zero if it fails, 1 if it succeeds
*
* CHANGES: lines
*************************************************************/
int delete(int beginLineIndex, int endLineIndex)
{
    // move all lines numbered higher than endLineIndex towards the beginning of lines
    if (beginLineIndex > endLineIndex || beginLineIndex < 0 || endLineIndex > lastLineIndex)
        return FALSE;
    moveDown(endLineIndex + 1, beginLineIndex, endLineIndex - beginLineIndex + 1);
    // adjust lastLineIndex
    lastLineIndex -= endLineIndex - beginLineIndex + 1;
    return TRUE;
}
/***********************************************************
* FUNCTION NAME: copy
*
* DESCRIPTION:  copies to scrap all lines from beginLineIndex to emdLineIndex, inclusive
*
* ARGUMENT LIST:
* Argument        Type         IO      Description
* ------------- --------       -- --------------------------------
* int           beginLineIndex  I  index of 1st line to delete
* int           endLineIndex    I  index of last line to delete
* RETURN VALUE: int  zero if it fails, 1 if it succeeds
*
* CHANGES: scrap
*************************************************************/
int copy(int beginLineIndex, int endLineIndex)
{
    if (endLineIndex > lastLineIndex || beginLineIndex < 0) return FALSE;
    int k = 0; // index in scrap

    for (int i = beginLineIndex; i <= endLineIndex; i++)
        strcpy(scrap[k++], lines[i]);
    lastScrapLine = endLineIndex - beginLineIndex;
    return TRUE;
}
/***********************************************************
* FUNCTION NAME: paste
*
* DESCRIPTION:  pastes scrap at current line index into lines
*
*
* ARGUMENT LIST: none
* RETURN VALUE: int
*
* CHANGES:
*************************************************************/
int paste()
{
    int success = FALSE;
    if (lastLineIndex + lastScrapLine + 1 >= MAX_LINE) return FALSE; // NO ROOM
    // make room for scrap at currentLineIndex in lines
    moveUp(crtLineIndex, crtLineIndex + lastScrapLine + 1, lastScrapLine + 1);
    // overwrite contents of that area in lines with contents of scrap
    for (int i = 0; i <= lastScrapLine; i++)
        strcpy(lines[crtLineIndex + i ], scrap[ i ]);
    lastLineIndex += lastScrapLine + 1;
    return success;
}
/***********************************************************
* FUNCTION NAME: cut
*
* DESCRIPTION:  cuts to scrap all lines from beginLineIndex to emdLineIndex, inclusive
*
* ARGUMENT LIST:
* Argument        Type         IO      Description
* ------------- --------       -- --------------------------------
* int           beginLineIndex  I  index of 1st line to cut
* int           endLineIndex    I  index of last line to cut
* RETURN VALUE: int  zero if it fails, 1 if it succeeds
*
* CHANGES: scrap, lines
*************************************************************/
int cut(int beginLineIndex, int endLineIndex)
{
    if (copy(beginLineIndex, endLineIndex))
        delete(beginLineIndex, endLineIndex);
    else
    {
        fprintf(stderr, "Cannot copy from line %d to line %d\n",
            beginLineIndex + 1, endLineIndex + 1);
        perror("");
        return FALSE;
    }
    return TRUE;
}
/***********************************************************
* FUNCTION NAME: view
*
* DESCRIPTION:  lists all lines from beginLineIndex to emdLineIndex, inclusive
*
* ARGUMENT LIST:
* Argument        Type         IO      Description
* ------------- --------       -- --------------------------------
* int           beginLineIndex  I  index of 1st line to list
* int           endLineIndex    I  index of last line to list
* RETURN VALUE: -
*
* CHANGES: -
*************************************************************/
void view(int beginLineIndex, int endLineIndex)
{
    int i, j;
    for (i = beginLineIndex, j = 0; i <= endLineIndex && i <= lastLineIndex; i++)
    {
        printf("%4d| %s\n", i + 1, lines[i]);
        if (i && i % 20 == 0)
        {
            printf("Press space to continue...");
            getchar();
        }
    }
}
/***********************************************************
* FUNCTION NAME: load
*
* DESCRIPTION:  loads file named fileName into lines
*
* ARGUMENT LIST:
* Argument        Type   IO      Description
* ------------- -------- -- --------------------------------
* const char*   fileName  I  name of file to load
* RETURN VALUE: int FALSE if cannot load, TRUE otherwise
*
* CHANGES: lines
*************************************************************/
int load(const char *fileName)
{
    FILE *f;
    static char buf[MAX_LINE_LEN];

    // open file to load
    if (NULL == (f = fopen(fileName, "r")))
    {   // open unsuccessful; inform user
        fprintf(stderr, "Cannot load %s\n", fileName);
        perror("");
        return FALSE;
    }
    else
    {  // open successful; read all contents into lines
        int i = 0;
        while (fgets(buf, sizeof(buf), f))
        {
            // remove ending newline
            if (buf[strlen(buf)-1] == '\n')
                buf[strlen(buf)-1] = '\0';
            strcpy(lines[i], buf);
            lastLineIndex = i;
            i++;
            memset(buf, '\0', sizeof(buf));
        }
        fclose(f);
    }
    return TRUE;
}
/***********************************************************
* FUNCTION NAME: write
*
* DESCRIPTION:  write lines into file named fileName
*
* ARGUMENT LIST:
* Argument        Type   IO      Description
* ------------- -------- -- --------------------------------
* const char*   fileName  I  name of file to save to
* RETURN VALUE: int FALSE if cannot load, TRUE otherwise
*
* CHANGES: -
*************************************************************/
int write(const char *fileName)
{
    FILE *f;
    static char buf[MAX_LINE_LEN];

    // open file to load
    if (NULL == (f = fopen(fileName, "w")))
    {   // open unsuccessful; inform user
        fprintf(stderr, "Cannot save to %s\n", fileName);
        perror("");
        return FALSE;
    }
    else
    {  // open successful; read all contents into lines
        int i;
        for (i = 0; i <= lastLineIndex; i++)
        {
            fputs(buf, f);
        }
        fclose(f);
    }
    return TRUE;
}


// end of ted.c
